@@ -15,7 +15,7 @@
 #' - PostgreSQL: ignored
 #' - DuckDB: ignored
 #'
-#' @return A data frame, one document per row
+#' @return Data frame, one document per row
 #'
 #' @export
 #'
@@ -43,6 +43,9 @@ docdb_get <- function(src, key, limit = NULL, ...) {
 #' @export
 docdb_get.src_couchdb <- function(src, key, limit = NULL, ...) {
 
+  # early return if container does not exist
+  if (!any(docdb_list(src) == key)) return(NULL)
+
   jsonlite::fromJSON(
     # get data
     sofa::db_alldocs(
@@ -63,6 +66,9 @@ docdb_get.src_couchdb <- function(src, key, limit = NULL, ...) {
 #' @export
 docdb_get.src_elastic <- function(src, key, limit = NULL, ...) {
 
+  # early return if container does not exist
+  if (!any(docdb_list(src) == key)) return(NULL)
+
   # adjust parameter
   if (is.null(limit)) limit <- 10000L
 
@@ -147,6 +153,9 @@ docdb_get.src_duckdb <- function(src, key, limit = NULL, ...) {
 #' @noRd
 sqlGet <- function(src, key, limit = NULL, getFunction, ...) {
 
+  # early return if container does not exist
+  if (!any(docdb_list(src) == key)) return(NULL)
+
   # set limit if not null
   n <- -1L
   if (!is.null(limit)) n <- limit
